/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "material.h"



Material::Material(std::string name, int numColor):
	mName(name), mNumColor(numColor),nbTexEmission(0), nbTexColor(0), nbTexGloss(0), nbTexNormal(0), materialKey(0),
	mActiveBlending(false), mBlendSrcFunc(GL_ONE), mBlendDstFunc(GL_ZERO), mAlphaTest(false), mAlphaTestValue(1.), mNbEntities(0) {
		
	mEmission = Color(numColor, -1., 1.);	
	mAmbientDiffuse = Color(numColor, 0., 1.);	
	mSpecular = Color(numColor, 0.5, 40.);	
}

Material::~Material() {
	for(int i=0;i<textures.size();i++) {
		delete textures[i];
	}
}


void Material::setName(std::string name) {
	mName = name;
}

std::string Material::getName() {
	return mName;
}


int Material::getNbEntities () {
	return mNbEntities;
}


void Material::setAmbientDiffuseColor(float coeff, float alpha) {
	mAmbientDiffuse = Color(mNumColor, coeff, alpha);
}

void Material::setAmbientDiffuseColor(Color color) {
	mAmbientDiffuse = color;
}

void Material::setSpecularColor(float coeff, float shininess) {
	mSpecular = Color(mNumColor, coeff, shininess);	
}

void Material::setSpecularColor(Color color) {
	mSpecular = color;
}

void Material::setEmissionColor(float coeff) {
	mEmission = Color(mNumColor, coeff, 0.);	
}

void Material::setEmissionColor(Color color) {
	mEmission = color;
}

	

Color Material::getSpecular() {
	return mSpecular;
}

Color Material::getAmbientDiffuse() {
	return mAmbientDiffuse;
}

Color Material::getEmission() {
	return mEmission;
}


int Material::getNbTexColor() {
	return nbTexColor;
}

int Material::getNbTexNormal() {
	return nbTexNormal;
}

int Material::getNbTexGloss() {
	return nbTexGloss;	
}

int Material::getNbTexEmi() {
	return nbTexEmission;	
}

int Material::getMaterialKey() {
	return materialKey;
}


void Material::setBlendFunc(GLenum blendSrcFunc, GLenum blendDstFunc) {
	mBlendSrcFunc = blendSrcFunc;
	mBlendDstFunc = blendDstFunc;
	
	mActiveBlending = (blendSrcFunc!=GL_ONE)||(blendDstFunc!=GL_ZERO);
}

void Material::setAlphaTest(float alphaTestValue) {
	mAlphaTestValue = alphaTestValue;
	mAlphaTest = (alphaTestValue > 0.0);
}


void Material::activate(int lightKey, shaderType shader) {
			
	if(mActiveBlending) {
		glEnable(GL_BLEND);
		glBlendFunc(mBlendSrcFunc,mBlendDstFunc);
	}
	
	if(mAlphaTest) {
		glEnable(GL_ALPHA_TEST);
		glAlphaFunc(GL_GREATER, mAlphaTestValue);
		glDisable(GL_CULL_FACE);
	}
	
	if(shader != FIXED_PIPELINE) {
		int shaderMaterialKey = shaderMana->getCurrentMaterialKey();
		int shaderLightKey = shaderMana->getCurrentLightKey();
		
		shaderType type = shaderMana->getCurrentShaderType();
		if((type!=shader)||(materialKey!=shaderMaterialKey)||(shaderLightKey!=lightKey)) {
			shaderMana->activateMaterialShader(lightKey, materialKey, PER_FRAG_LIGHT);	
		}		

		for(int i=0;i<textures.size();i++){
			glActiveTexture(GL_TEXTURE0 + i);
			textures[i]->activate();
			int color = -1;
			int normal = -1;
			int gloss = -1;
			int emi = -1;
			
			switch(textures[i]->getType()) {				
				case COLOR_MAP:
					color = i;
					break;

				case NORMAL_MAP:
					normal = i;
					break;					
					
				case GLOSS_MAP:
					gloss = i;
					break;
				
				case EMISSION_MAP:
					emi = i;
					break;					
			}
			shaderMana->setNumMaps(color, normal, gloss, emi);
		}		
	} else {
		for(int i=0;i<textures.size();i++){
			glActiveTexture(GL_TEXTURE0 + i);
			textures[i]->activate();	
		}
	}
	
	glActiveTexture(GL_TEXTURE0);	
	
	if(nbTexColor == 0) {
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mAmbientDiffuse);
	} else {
		glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, Color::WHITE);
	}
	
	
	if(nbTexEmission == 0) {
		glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, mEmission);			
	}		
	
	if(nbTexGloss == 0) {
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mSpecular.a);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mSpecular);
	}
	
}

void Material::release() {
	
	if(mActiveBlending) {
		glDisable(GL_BLEND);
	}
	
	if(mAlphaTest) {
		glDisable(GL_ALPHA_TEST);
		glEnable(GL_CULL_FACE);
	}	
	
	for(int i=0;i<textures.size();i++){
	    glActiveTexture(GL_TEXTURE0 + i);
	    textures[i]->release();
        }
	glActiveTexture(GL_TEXTURE0);		
}



void Material::addTexture(wstring text, int color, textureType type, int width, int height, GLenum wrapMode, bool mipmaps) {
	
	bool canAddTexture = ((nbTexColor + nbTexNormal + nbTexGloss + nbTexEmission) < getMaxTextureUnits());

	if(canAddTexture) {
		ImagePng *image = new ImagePng(text, color, width, height);
		
		if((image->getWidth() > getTextureSizeMax())&&(image->getHeight() > getTextureSizeMax())) {
			//texture size is not supported by the hardware
			delete image;
		} else {
			addTextureAndUpdateNbTex(image, type, wrapMode, mipmaps, 1., 1.);			
		}
	}
}

void Material::addTexture(wstring text, int textColor, float textColorCoeff, int backgroundColor, float backgroundCoeff, textureType type, int width, int height, GLenum wrapMode, bool mipmaps) {
	
	bool canAddTexture = ((nbTexColor + nbTexNormal + nbTexGloss + nbTexEmission) < getMaxTextureUnits());

	if(canAddTexture) {
		ImagePng *image = new ImagePng(text, textColor, textColorCoeff, backgroundColor, backgroundCoeff, width, height);
		
		if((image->getWidth() > getTextureSizeMax())&&(image->getHeight() > getTextureSizeMax())) {
			//texture size is not supported by the hardware
			delete image;
		} else {
			addTextureAndUpdateNbTex(image, type, wrapMode, mipmaps, 1., 1.);			
		}
	}	
}



void Material::addTexture(std::string filename, bool smallTex, int color, textureType type, GLenum wrapMode, bool mipmaps, GraphicOptions *graphics) {
	addTexture(filename, smallTex, color, type, wrapMode, 1., 1., mipmaps, graphics);
}

void Material::addTexture(std::string filename, bool smallTex, int color, textureType type, GLenum wrapMode, float scaleS, float scaleT, bool mipmaps, GraphicOptions *graphics) {

	bool canAddTexture = testNbTextures(type, graphics);
	
	std::string filepath;
	
	if(canAddTexture) {
		if(smallTex) {
			filepath = fileMana->getSmallTexturesDir() + "/" + filename;
		} else {
			if(graphics->getTextureMode() == TEXTURE_LOW) {
				filepath = fileMana->getLowTexturesDir() + "/" + filename;
			} else {	//TEXTURE_HIGH
				filepath = fileMana->getHightTexturesDir() + "/" + filename;
			}
		}
		
		ImagePng *image;
		
		if(fileExists(filepath)) {
			image = new ImagePng(filepath, color);
			
			//is texture size supported by the hardware ?
			canAddTexture = ((image->getWidth() <= getTextureSizeMax())&&(image->getHeight() <= getTextureSizeMax()));
			
			if(!canAddTexture) delete image;
		} else {
			canAddTexture = false;
		}
		
		if((!canAddTexture)&&(graphics->getTextureMode() == TEXTURE_HIGH)) {
			//we try the low resolution texture
			filepath = fileMana->getLowTexturesDir() + "/" + filename;
			
			if(fileExists(filepath)) {
				image = new ImagePng(filepath, color);
				
				//is texture size supported by the hardware ?
				canAddTexture = ((image->getWidth() <= getTextureSizeMax())&&(image->getHeight() <= getTextureSizeMax()));				
			
				if(!canAddTexture) delete image;
			} else {
				canAddTexture = false;
			}
		}
		
		if(canAddTexture) {
			addTextureAndUpdateNbTex(image, type, wrapMode, mipmaps, scaleS, scaleT);
		}
	}	
	
}

int Material::getNumberOfTextures() {
	return textures.size();
}


void Material::incNbEntities() {
	mNbEntities++;
}

void Material::decNbEntities() {
	mNbEntities--;
}




bool Material::testNbTextures(textureType type, GraphicOptions *graphics) {
	bool canAddTexture = ((nbTexColor + nbTexNormal + nbTexGloss + nbTexEmission) < getMaxTextureUnits());
	
	if(canAddTexture) {
		if(graphics == NULL) {
			canAddTexture == true;
		} else {
			if(graphics->getTextureMode() != TEXTURE_NONE) {
				canAddTexture = (((type == COLOR_MAP)&&((graphics->getTextureMode() == TEXTURE_HIGH)||(nbTexColor < 2)))||
							  ((type == NORMAL_MAP)&&(graphics->areNormalMapsEnabled())&&((graphics->getTextureMode() == TEXTURE_HIGH)||(nbTexNormal == 0)))||
							  ((type == GLOSS_MAP)&&(graphics->areGlossMapsEnabled())&&((graphics->getTextureMode() == TEXTURE_HIGH)||(nbTexGloss == 0)))||
							  ((type == EMISSION_MAP)&&(graphics->areEmiMapsEnabled())&&((graphics->getTextureMode() == TEXTURE_HIGH)||(nbTexEmission == 0)))||
							  (type == OTHER_MAP));
			} else {
				canAddTexture = false;
			}
		}
	}
	
	return canAddTexture;
}

void Material::addTextureAndUpdateNbTex(ImagePng *image, textureType type, GLenum wrapMode, bool mipmaps, float scaleS, float scaleT) {
	TextureLayer *texture = new TextureLayer (image, type, wrapMode, mipmaps);
	texture->setScale(scaleS, scaleT);
	textures.push_back(texture);
	delete image;
	
	switch(type) {
		case COLOR_MAP:
			nbTexColor++;
			materialKey++;
			break;
			
		case NORMAL_MAP:
			nbTexNormal++;
			materialKey+=4;	
			break;				
			
		case GLOSS_MAP:
			nbTexGloss++;
			materialKey+=16;
			break;
		
		case EMISSION_MAP:
			nbTexEmission++;
			materialKey+=64;
			break;			
			
		default:
			break;	
	}	
}

